/*
**      Maximus 3.0 to Maximus 2.0 Compatibility Fixes for SILT(P)
**      Public Domain --- Use and hack on freely.
**      Written by George Hatchew
**
**
**  Module name:    FIXSILT.C
**  Description:    Program to read through Max 3.0 MAREA.DAT, converting all
**                  AKA offsets to the Max 2.0 AREA.DAT equivalent.
**
*/

#include <ctype.h>
#include <fcntl.h>
#include <io.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>

#include "fixsilt.h"

/*--------------------------------------------------------------------------*/

/*
**  Global variables
*/

char *offsets = NULL;               /* Maximus .PRM 'heap-o-strings'        */
struct m_pointers maxprm;           /* Maximus .PRM fixed-length structure  */

/*--------------------------------------------------------------------------*/

/*
**  Load the Maximus .PRM file into memory.  Return TRUE on a successful
**  load, FALSE otherwise.
*/

static int LoadMaxPrmFile(char *prmname)
{
    int prmfile, x;

    if ((prmfile = open(prmname, O_BINARY|O_RDONLY)) == -1)
    {
        fprintf(stdout, "\nUnable to open %s!\n", prmname);
        return(FALSE);
    }

    read(prmfile, &maxprm, sizeof(struct m_pointers));
    if ((maxprm.id != 'M') || (maxprm.version != CTL_VER))
    {
        fputs("\nThis does not appear to be a Maximus 2.x or 3.x .PRM file.\n",
            stdout);
        return(FALSE);
    }

    lseek(prmfile, (long) maxprm.heap_offset, SEEK_SET);
    x = (int) (filelength(prmfile) - ((int) tell(prmfile)));
    if ((offsets = (char *) malloc(x)) == NULL)
    {
        fputs("\nNot enough memory!\n", stdout);
        return(FALSE);
    }

    read(prmfile, offsets, x);
    close(prmfile);
    return(TRUE);
}

/*---------------------------------------------------------------------------*/

/*
**  Stolen from the Squish code, this function simply returns the 'SquishHash'
**  of a string.  A convenient way of representing 'strings' as a 4-byte
**  number.
*/

static DwdT SquishHash(char *areaname)
{
    char *p;
    DwdT hash = 0, g;

    for (p = areaname; *p != '\0'; p++)
    {
        hash = (hash << 4) + (DwdT) tolower(*p);

        if ((g = (hash & 0xf0000000L)) != 0L)
        {
            hash |= (g >> 24);
            hash |= g;
        }
    }
  
    return (hash & 0x7fffffffLu);
}

/*--------------------------------------------------------------------------*/

/*
**  Locate the current message area in the old AREA.DAT file (by looking
**  through the hash indexes).  This function simply seeks to the proper
**  record in the old AREA.DAT file for updating.  Returns TRUE if the old
**  area could be found, or FALSE if the area couldn't be located (!).
*/

static int SeekToOldArea(int adatfile, char *descript, char *echotag,
                         DwdT *dschash, DwdT *taghash)
{
    int i = 0;
    char tempbuf[80];
    long seekpos;
    DwdT dscnew, tagnew;

    /*
    **  We only want to get a SquishHash of
    **  the number of characters that would fit
    **  into the old structures.  (MAREA.DAT allows
    **  for 'strings' up to 65,535 bytes in length.)
    */

    strncpy(tempbuf, descript, 79);
    tempbuf[79] = '\0';
    dscnew = SquishHash(tempbuf);

    strncpy(tempbuf, echotag, 39);
    tempbuf[39] = '\0';
    tagnew = SquishHash(tempbuf);

    /*
    **  If both hashes turned out to be 0L, we
    **  don't really have a valid message area.
    */

    if ((dscnew != 0L) && (tagnew != 0L))
    {
        while (dschash[i] != 0xFFFFFFFFL)
        {
            if ((dscnew == dschash[i]) && (tagnew == taghash[i]))
                break;
            i++;
        }
        if (dschash[i] != 0xFFFFFFFFL)
        {
            /*
            **  Yeah, I know that AREA.DAT had a variable-length record
            **  size, but SILT 3.0 will always generate record sizes
            **  that are sizeof(OLD_AREADAT).
            */

            seekpos = (long) ((long) i * (long) sizeof(OLD_AREADAT));
            return (lseek(adatfile, seekpos, SEEK_SET) == seekpos);
        }
    }
    return(FALSE);
}


/*--------------------------------------------------------------------------*/

/*
**  Convert to the old-style AREA.DAT file from the information in MAX.PRM
**  and MAREA.DAT.
*/

static int ConvertNewToOld(int mareafile, int adatfile, DwdT *dschash,
                           DwdT *taghash)
{
    int done = FALSE, i = 0;
    char *heap;
    MAREA newarea;
    OLD_AREADAT oldarea;

    fputs("Loading areas...", stdout);
    fflush(stdout);
    lseek(adatfile, 0L, SEEK_SET);
    while (read(adatfile, &oldarea, sizeof(OLD_AREADAT)))
    {
        dschash[i] = SquishHash(oldarea.msginfo);
        taghash[i++] = SquishHash(oldarea.msgname);
    }
    dschash[i] = 0xFFFFFFFFL;
    taghash[i] = 0xFFFFFFFFL;

    fputs("\rWriting areas...", stdout);
    fflush(stdout);
    lseek(mareafile, 4L, SEEK_SET);
    while (!done)
    {
        memset(&newarea, 0, sizeof(MAREA));
        if (read(mareafile, &newarea, sizeof(MAREA)) != sizeof(MAREA))
        {
            done = TRUE;
            continue;
        }

        if ((newarea.cbHeap > 0) &&
            ((heap = (char *) malloc(newarea.cbHeap)) != NULL))
        {
            if (newarea.num_override > 0)
            {
                lseek(mareafile, (long) (newarea.num_override * sizeof(OVERRIDE_MENU)),
                    SEEK_CUR);
            }

            memset(heap, 0, newarea.cbHeap);
            read(mareafile, heap, newarea.cbHeap);

            for (i = 0; i < ALIAS_CNT; i++)
            {
                if ((newarea.primary.zone == maxprm.address[i].zone) &&
                    (newarea.primary.net == maxprm.address[i].net) &&
                    (newarea.primary.node == maxprm.address[i].node) &&
                    (newarea.primary.point == maxprm.address[i].point))
                {
                    break;
                }
            }

            if (i >= ALIAS_CNT)
                i = 0;

            if (!(newarea.attribs & MA_DIVBEGIN) &&
                !(newarea.attribs & MA_DIVEND))
            {
                if (SeekToOldArea(adatfile, MAREA_STRING(newarea, descript),
                    MAREA_STRING(newarea, echo_tag), dschash, taghash))
                {
                    read(adatfile, &oldarea, sizeof(OLD_AREADAT));
                    oldarea.origin_aka = (IntT) i;

                    lseek(adatfile, - ((long) sizeof(OLD_AREADAT)), SEEK_CUR);
                    write(adatfile, &oldarea, sizeof(OLD_AREADAT));
                }
            }
        }
    }
    fputs("\n", stdout);
    return(TRUE);
}

/*--------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
    int adatfile = -1, mareafile = -1, numrecs, errlvl = 0;
    char tempbuf[256];
    DwdT *dschash = NULL, *taghash = NULL;

    fputs("\nFIXSILT v1.0  Fix Maximus 2.x AREA.DAT files after running SILT 3.0\n",
        stdout);

    if (argc > 1)
        strcpy(tempbuf, argv[1]);
    else
        strcpy(tempbuf, "MAX.PRM");

    /*
    **  Load the Maximus PRM file.
    */

    if (!LoadMaxPrmFile(tempbuf))
        return(1);

    /*
    **  Open the old AREA.DAT file.
    */

    strupr(PRM(adat_name));
    if ((adatfile = open(PRM(adat_name), O_BINARY|O_RDWR)) == -1)
    {
        fprintf(stdout, "\nUnable to open %s.  Run SILT(P) with the -2s option.\n",
            PRM(adat_name));
        errlvl = 1;
    }

    /*
    **  Open the new MAREA.DAT file.
    */

    if (!errlvl)
    {
        strcpy(tempbuf, strupr(PRM(marea_name)));
        strcat(tempbuf, ".DAT");
        if ((mareafile = open(tempbuf, O_BINARY|O_RDONLY)) == -1)
        {
            fprintf(stdout, "\nUnable to open %s.  Run SILT(P) with the -2s option.\n",
                tempbuf);
            errlvl = 1;
        }
    }

    /*
    **  Allocate memory for the hash indices.
    */

    if (!errlvl)
    {
        numrecs = (int) (filelength(adatfile) / (long) sizeof(OLD_AREADAT) + 1);
        if ((dschash = (DwdT *) calloc(numrecs, sizeof(DwdT))) == NULL)
        {
            fputs("Not enough memory.\n", stdout);
            errlvl = 1;
        }

        if ((taghash = (DwdT *) calloc(numrecs, sizeof(DwdT))) == NULL)
        {
            fputs("Not enough memory.\n", stdout);
            errlvl = 1;
        }
    }

    /*
    **  Fix the old-style file.
    */

    if (!errlvl)
        ConvertNewToOld(mareafile, adatfile, dschash, taghash);

    /*
    **  Clean up the mess we left behind.
    */

    if (dschash != NULL)
        free(dschash);
    if (taghash != NULL)
        free(taghash);
    if (offsets != NULL)
        free(offsets);
    if (adatfile != -1)
        close(adatfile);
    if (mareafile != -1)
        close(mareafile);

    fputs("Done!\n", stdout);

    return(errlvl);
}

